from typing import Optional

from edc_reportable import GRADE3, GRADE4


class AlreadyRegistered(Exception):
    pass


class ReferenceRangeCollectionError(Exception):
    pass


class ReferenceRangeCollection:

    """Holds all normal and grading data available by name.

    Usually there is just one reference collection per project so
    name can be the project name.
    """

    def __init__(
        self,
        name=None,
        reportable_grades: Optional[list] = None,
        reportable_grades_exceptions: Optional[dict] = None,
    ):
        self.registry = {}
        self.name = name
        self.reportable_grades = reportable_grades or [GRADE3, GRADE4]
        self.reportable_grades_exceptions = reportable_grades_exceptions or {}
        for k in self.reportable_grades_exceptions:
            if k not in self.reportable_grades:
                raise ReferenceRangeCollectionError(
                    "Invalid reportable grades exception. UTEST_ID not found "
                    f"in grading table. Got {k}"
                )

    def __repr__(self):
        return f"{self.__class__.__name__}('{self.name}')"

    def register(self, grp=None):
        if grp.name in self.registry:
            raise AlreadyRegistered(f"Got {repr(grp)}")
        else:
            self.registry.update({grp.name: grp})

    def get(self, name):
        return self.registry.get(name)

    def update_grp(self, grp):
        self.registry.update({grp.name: grp})

    def as_data(self):
        """Returns a dictionary of the normal and grading references
        in this collection.
        """
        exclude_attrs = [
            "evaluator",
            "age_evaluator",
            "_normal_reference",
            "normal_references",
        ]
        data = {"normal": [], "grading": []}
        for name, grp in self.registry.items():
            for normal_refs in grp.normal.values():
                for ref in normal_refs:
                    dct = ref.__dict__
                    dct.update(name=name)
                    dct = {k: v for k, v in dct.items() if k not in exclude_attrs}
                    data["normal"].append(dct)
        for name, grp in self.registry.items():
            for grade_refs in grp.grading.values():
                for ref in grade_refs:
                    dct = ref.__dict__
                    dct.update(name=name)
                    dct = {k: v for k, v in dct.items() if k not in exclude_attrs}
                    data["grading"].append(dct)
        return data



from edc_constants.constants import FEMALE

from ...units import MICROMOLES_PER_LITER
from .base_egrfr import BaseEgfr


class EgfrCockcroftGault(BaseEgfr):

    """Reference https://www.mdcalc.com/creatinine-clearance-cockcroft-gault-equation

    Cockcroft-Gault

    eGFR (mL/min) = { (140 – age (years)) x weight (kg) x constant*} /
    serum creatinine (μmol/L)
    *constant = 1.23 for males and 1.05 for females

    Cockcroft-Gault CrCl, mL/min =
        (140 – age) × (weight, kg) × (0.85 if female) / (72 × SCr(mg/dL))

    """

    @property
    def value(self):
        if (
            self.gender
            and self.age_in_years
            and self.weight
            and self.scr.get(MICROMOLES_PER_LITER)
        ):
            return (self.age_factor * self.weight * self.gender_factor) / (
                self.scr.get(MICROMOLES_PER_LITER)
            )
        return None

    @property
    def gender_factor(self):
        return 1.05 if self.gender == FEMALE else 1.23
        # return 0.85 if self.gender == FEMALE else 1.00

    @property
    def age_factor(self):
        return 140 - self.age_in_years



from .age_evaluator import AgeEvaluator
from .evaluator import Evaluator, ValueBoundryError
from .parsers import parse_boundary


class ValueReferenceError(Exception):
    pass


class ValueReference:

    age_evaluator_cls = AgeEvaluator
    evaluator_cls = Evaluator

    def __init__(self, name=None, gender=None, units=None, normal_references=None, **kwargs):
        self._normal_reference = None
        self.normal_references = normal_references
        self.name = name
        self.units = units
        if isinstance(gender, (list, tuple)):
            self.gender = "".join(gender)
        else:
            self.gender = gender
        kwargs["lower"] = self.get_boundary_value(kwargs["lower"])
        kwargs["upper"] = self.get_boundary_value(kwargs["upper"])
        self.evaluator = self.evaluator_cls(name=self.name, units=units, **kwargs)
        self.age_evaluator = self.age_evaluator_cls(**kwargs)
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name}, {self.description()})"

    def description(self, **kwargs):
        return (
            f"{self.evaluator.description(**kwargs)} {self.gender} "
            f"{self.age_evaluator.description()}"
        )

    def in_bounds(self, value=None, **kwargs):
        try:
            in_bounds = self.evaluator.in_bounds_or_raise(value, **kwargs)
        except ValueBoundryError:
            in_bounds = False
        return in_bounds

    def age_match(self, dob=None, report_datetime=None, age_units=None):
        try:
            age_match = self.age_evaluator.in_bounds_or_raise(
                dob=dob, report_datetime=report_datetime, age_units=age_units
            )
        except ValueBoundryError:
            age_match = False
        return age_match

    @property
    def normal_reference(self):
        if not self._normal_reference:
            if self.normal_references:
                self._normal_reference = [
                    x[0] for x in self.normal_references.values() if x[0].units == self.units
                ]
            else:
                raise ValueReferenceError(
                    f"Normal references not provided. Got {self.name} per {self.units}"
                )
            if not self._normal_reference:
                opts = [(x[0].name, x[0].units) for x in self.normal_references.values()]
                raise ValueReferenceError(
                    "Normal reference not found. Expected one "
                    f"of {opts}. Got {self.name} per {self.units}"
                )
        return self._normal_reference[0]

    def get_boundary_value(self, value):
        """Return value as a literal value or as a value relative
        to the normal lower or upper normal.
        """
        try:
            value = value.upper()
        except AttributeError:
            pass
        else:
            lln, uln = parse_boundary(value)
            value = (
                lln * self.normal_reference.lower if lln else uln * self.normal_reference.upper
            )
        return value



from .constants import GRADE0, GRADE1, GRADE2, GRADE3, GRADE4, GRADE5
from .value_reference import ValueReference


class GradeError(Exception):
    pass


class GradeReference(ValueReference):

    grades = [GRADE0, GRADE1, GRADE2, GRADE3, GRADE4, GRADE5]

    def __init__(self, grade=None, normal_references=None, **kwargs):
        if int(grade) not in [int(x) for x in self.grades]:
            raise GradeError(f"Invalid grade. Expected one of {self.grades}. Got {grade}.")
        self.grade = int(grade)
        super().__init__(normal_references=normal_references, **kwargs)

    def __repr__(self):
        return f"{super().__repr__()} GRADE {self.grade})"

    def description(self, **kwargs):
        return f"{self.evaluator.description(**kwargs)} GRADE {self.grade}"



from .bmi_form_validator_mixin import BmiFormValidatorMixin
from .egfr_form_validator_mixins import (
    EgfrCkdEpiFormValidatorMixin,
    EgfrCockcroftGaultFormValidatorMixin,
)
from .reportables_form_validator_mixin import ReportablesFormValidatorMixin



import copy
import csv
import os
import sys
from importlib import import_module
from typing import Optional

from django.apps import apps as django_apps
from django.core.management.color import color_style
from django.utils.module_loading import module_has_submodule

from .grade_reference import GradeReference
from .normal_reference import NormalReference
from .parsers import unparse
from .reference_range_collection import ReferenceRangeCollection
from .value_reference_group import GRADING, NORMAL, ValueReferenceGroup


class SiteReportablesError(Exception):
    pass


class MissingNormalReference(Exception):
    pass


class SiteReportables:
    def __init__(self):
        self._registry = {}

    def __iter__(self):
        return iter(self._registry.items())

    def register(
        self,
        name: str = Optional[None],
        normal_data: Optional[dict] = None,
        grading_data: Optional[dict] = None,
        reportable_grades: Optional[list] = None,
        reportable_grades_exceptions: Optional[dict] = None,
    ):
        if name in self._registry:
            reference_range_collection = self._registry.get(name)
        else:
            reference_range_collection = ReferenceRangeCollection(
                name=name,
                reportable_grades=reportable_grades,
                reportable_grades_exceptions=reportable_grades_exceptions,
            )
        for name, datas in normal_data.items():
            grp = ValueReferenceGroup(name=name)
            for data in datas:
                val_ref = NormalReference(name=name, **data)
                grp.add_normal(val_ref)
            reference_range_collection.register(grp)
        for name, datas in grading_data.items():
            grp = reference_range_collection.get(name)
            if not grp:
                raise MissingNormalReference(
                    f"Attempting to add grading for item without a "
                    f"normal reference. Got {name}."
                )
            for data in datas:
                grade_ref = GradeReference(name=name, normal_references=grp.normal, **data)
                grp.add_grading(grade_ref)
            reference_range_collection.update_grp(grp)
        site_reportables._registry.update(
            {reference_range_collection.name: reference_range_collection}
        )

    def get(self, name):
        return self._registry.get(name)

    def get_normal(self, name):
        return self._registry.get(name)[NORMAL]

    def get_grading(self, name):
        return self._registry.get(name)[GRADING]

    def get_reportable_grades(self, name):
        return self._registry.get(name).reportable_grades

    def read_csv(self, name=None, path=None):
        pass

    def to_csv(self, collection_name=None, path=None):
        path = path or "~/"
        path = os.path.expanduser(path)
        filename1 = os.path.join(path, f"{collection_name}_normal_ranges.csv")
        filename2 = os.path.join(path, f"{collection_name}_grading.csv")
        reference_range_collection = self.get(collection_name)
        data = reference_range_collection.as_data()
        try:
            fieldnames = list(data.get(NORMAL)[0].keys())
        except IndexError:
            pass
        else:
            fieldnames.insert(1, "description")
            with open(filename1, "w") as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()
                for dct in data.get(NORMAL):
                    dct.update(description=unparse(**dct))
                    writer.writerow(dct)
        try:
            fieldnames = list(data.get(GRADING)[0].keys())
        except IndexError:
            pass
        else:
            fieldnames.insert(1, "description")
            with open(filename2, "w") as f:
                writer = csv.DictWriter(f, fieldnames=fieldnames)
                writer.writeheader()
                for dct in data.get(GRADING):
                    dct.update(description=unparse(**dct))
                    writer.writerow(dct)
        return filename1, filename2

    def autodiscover(self, module_name=None, verbose=True):
        module_name = module_name or "reportables"
        writer = sys.stdout.write if verbose else lambda x: x
        style = color_style()
        writer(f" * checking for site {module_name} ...\n")
        for app in django_apps.app_configs:
            writer(f" * searching {app}           \r")
            try:
                mod = import_module(app)
                try:
                    before_import_registry = copy.copy(site_reportables._registry)
                    import_module(f"{app}.{module_name}")
                    writer(f" * registered '{module_name}' from '{app}'\n")
                except SiteReportablesError as e:
                    writer(f"   - loading {app}.{module_name} ... ")
                    writer(style.ERROR(f"ERROR! {e}\n"))
                except ImportError as e:
                    site_reportables._registry = before_import_registry
                    if module_has_submodule(mod, module_name):
                        raise SiteReportablesError(str(e))
            except ImportError:
                pass
            except Exception as e:
                raise SiteReportablesError(
                    f"{e.__class__.__name__} was raised when loading {module_name}. "
                    f"Got {e} See {app}.{module_name}"
                )


site_reportables = SiteReportables()



import os
from tempfile import mkdtemp

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import MALE
from edc_utils import get_utcnow

from edc_reportable import IU_LITER, site_reportables
from edc_reportable.grading_data.daids_july_2017 import chemistries
from edc_reportable.normal_data.africa import normal_data as africa_normal_data
from reportable_app.reportables import grading_data, normal_data


class TestSiteReportables(TestCase):
    def setUp(self):
        site_reportables._registry = {}

        site_reportables.register(
            name="my_reference_list", normal_data=normal_data, grading_data=grading_data
        )

    def test_to_csv(self):
        path = mkdtemp()
        filename1, filename2 = site_reportables.to_csv(
            collection_name="my_reference_list", path=path
        )
        with open(os.path.join(path, filename1)) as f:
            header = str(f.readline()).strip()
            self.assertEqual(
                header,
                (
                    "name,description,units,gender,lower,upper,lower_inclusive,"
                    "upper_inclusive,fasting,age_lower,age_upper,"
                    "age_units,age_lower_inclusive"
                ),
            )

    def test_haemoglobin(self):
        reportables = site_reportables.get("my_reference_list")
        haemoglobin = reportables.get("haemoglobin")
        normal = haemoglobin.get_normal(
            value=15.0,
            units="g/dL",
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertIsNotNone(normal)
        self.assertIn("13.5<=15.0<=17.5", normal.description)

        grade = haemoglobin.get_grade(
            value=8,
            units="g/dL",
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertIn("7.0<=8.0<9.0", grade.description)

        grade = haemoglobin.get_grade(
            value=15,
            units="g/dL",
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertIsNone(grade)


class TestSiteReportablesDiads2017(TestCase):
    def setUp(self):
        site_reportables._registry = {}
        grading_data.update(**chemistries)
        site_reportables.register(
            name="my_reference_list", normal_data=africa_normal_data, grading_data=grading_data
        )

    def test_daids_2017(self):
        reportables = site_reportables.get("my_reference_list")
        amylase = reportables.get("amylase")
        normal = amylase.get_normal(
            value=50,
            units=IU_LITER,
            gender=MALE,
            dob=get_utcnow() - relativedelta(years=25),
        )
        self.assertIsNotNone(normal)
        self.assertIn("25.0<=50.0<=125.0", normal.description)



from edc_constants.constants import FEMALE, MALE

from edc_reportable import (
    GRAMS_PER_DECILITER,
    IU_LITER,
    MICROMOLES_PER_LITER,
    MILLIGRAMS_PER_DECILITER,
    MILLIMOLES_PER_LITER,
    TEN_X_9_PER_LITER,
    adult_age_options,
)
from edc_reportable import dummy_parse as dummy
from edc_reportable import parse as p
from edc_reportable.units import (
    CELLS_PER_MILLIMETER_CUBED,
    EGFR_UNITS,
    GRAMS_PER_LITER,
    PERCENT,
    PLUS,
)

normal_data = {
    "albumin": [
        p(
            "3.5<=x<=5.0",
            units=GRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        p(
            "35<=x<=50",
            units=GRAMS_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "alp": [p("40<=x<=150", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "alt": [p("0<=x<=55", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "amylase": [p("25<=x<=125", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "ast": [p("5<=x<=34", units=IU_LITER, gender=[MALE, FEMALE], **adult_age_options)],
    "egfr": [dummy(units=EGFR_UNITS)],
    "chol": [
        p(
            "0.5<=x<=6.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "creatinine": [
        p(
            "63.6<=x<=110.5",
            units=MICROMOLES_PER_LITER,
            gender=[MALE],
            **adult_age_options,
        ),
        p(
            "50.4<=x<=98.1",
            units=MICROMOLES_PER_LITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
        p(
            "0.6<=x<=1.3",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "crp": [
        p("0.0<=x<1.0", units=PLUS, gender=[MALE, FEMALE], **adult_age_options),
    ],
    "glucose": [
        p(
            "4.0<=x<=6.11",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            fasting=True,
            **adult_age_options,
        ),
        p(
            "4.0<=x<=6.44",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            fasting=False,
            **adult_age_options,
        ),
    ],
    "hba1c": [p("4.4<=x<=6.6", units=PERCENT, gender=[MALE, FEMALE], **adult_age_options)],
    "hdl": [
        p(
            "1.04<=x<=1.55",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "ggt": [
        p("12<=x<=64", units=IU_LITER, gender=[MALE], **adult_age_options),
        p("9<=x<=36", units=IU_LITER, gender=[FEMALE], **adult_age_options),
    ],
    "haemoglobin": [
        p(
            "13.0<=x<=17.0",
            units=GRAMS_PER_DECILITER,
            gender=[MALE],
            **adult_age_options,
        ),
        p(
            "12.0<=x<=15.0",
            units=GRAMS_PER_DECILITER,
            gender=[FEMALE],
            **adult_age_options,
        ),
    ],
    # hematocrit
    "hct": [p("37.0<=x<=54.0", units=PERCENT, gender=[MALE, FEMALE], **adult_age_options)],
    "ldl": [
        p(
            "0.00<=x<=3.34",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "magnesium": [
        p(
            "0.75<=x<=1.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        p(
            "1.8<=x<=2.9",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "neutrophil": [
        p(
            "2.5<=x<=7.5",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "platelets": [
        p(
            "150<=x<=450",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        p(
            "150000<=x<=450000",
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "potassium": [
        p(
            "3.6<=x<=5.2",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "proteinuria": [
        p(
            "0.0<=x<1.0",
            units=PLUS,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "sodium": [
        p(
            "135<=x<=145",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "trig": [
        p(
            "0.00<=x<=1.69",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    # BUN
    "urea": [
        p(
            "2.5<=x<=6.5",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        )
    ],
    "uric_acid": [
        p(
            "0.15<=x<=0.35",
            units=MILLIMOLES_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        p(
            "7.2<=x",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "rbc": [
        p(
            "3.5<=x<=5.5",
            units=TEN_X_9_PER_LITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
        p(
            "3500<=x<=5500",
            units=CELLS_PER_MILLIMETER_CUBED,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    # TODO: tbil normal range is not set. range needed
    "tbil": [
        p(
            "0.0<=x<9999.0",
            units=MILLIGRAMS_PER_DECILITER,
            gender=[MALE, FEMALE],
            **adult_age_options,
        ),
    ],
    "wbc": [
        p("2.49<x", units=TEN_X_9_PER_LITER, gender=[MALE, FEMALE], **adult_age_options),
    ],
}



from copy import copy

from django.apps import apps as django_apps
from edc_constants.constants import YES

from ..reportables_evaluator import ReportablesEvaluator


class ReportablesFormValidatorMixin:

    reportables_cls = ReportablesEvaluator
    value_field_suffix = "_value"

    @property
    def reportables_evaluator_options(self):
        return {}

    def validate_reportable_fields(
        self, reference_range_collection_name: str, **reportables_evaluator_options
    ):
        """Called in clean() method of the FormValidator.

        for example:

            def clean(self):
                ...
                self.validate_reportable_fields()
                ...
        """

        subject_visit = self.cleaned_data.get("subject_visit")
        registered_subject_model_cls = django_apps.get_model(
            "edc_registration.registeredsubject"
        )
        subject_identifier = self.cleaned_data.get("subject_visit").subject_identifier
        registered_subject = registered_subject_model_cls.objects.get(
            subject_identifier=subject_identifier
        )

        # check normal ranges and grade result values
        reportables = self.reportables_cls(
            reference_range_collection_name,
            cleaned_data=copy(self.cleaned_data),
            gender=registered_subject.gender,
            dob=registered_subject.dob,
            report_datetime=subject_visit.report_datetime,
            value_field_suffix=self.value_field_suffix,
            **reportables_evaluator_options,
        )
        reportables.validate_reportable_fields()

        reportables.validate_results_abnormal_field()
        self.applicable_if(
            YES, field="results_abnormal", field_applicable="results_reportable"
        )
        reportables.validate_results_reportable_field()



from typing import Optional

from edc_constants.constants import BLACK, FEMALE

# TODO: https://www.rcpa.edu.au/Manuals/RCPA-Manual/
#  Pathology-Tests/C/Creatinine-clearance-Cockcroft-and-Gault
from ...units import MILLIGRAMS_PER_DECILITER
from .base_egrfr import BaseEgfr


class EgfrCkdEpi(BaseEgfr):  # noqa

    """Reference https://nephron.com/epi_equation

    CKD-EPI Creatinine equation

    Levey AS, Stevens LA, et al. A New Equation to Estimate Glomerular
    Filtration Rate. Ann Intern Med. 2009; 150:604-612.
    """

    @property
    def value(self) -> Optional[float]:
        if (
            self.gender
            and self.age_in_years
            and self.ethnicity
            and self.scr.get(MILLIGRAMS_PER_DECILITER)
        ):
            scr = self.scr.get(MILLIGRAMS_PER_DECILITER)
            return (
                141.000
                * (min(scr / self.kappa, 1.000) ** self.alpha)
                * (max(scr / self.kappa, 1.000) ** -1.209)
                * self.age_factor
                * self.gender_factor
                * self.ethnicity_factor
            )
        return None

    @property
    def alpha(self) -> float:
        return -0.329 if self.gender == FEMALE else -0.411

    @property
    def kappa(self) -> float:
        return 0.7 if self.gender == FEMALE else 0.9

    @property
    def ethnicity_factor(self) -> float:
        return 1.150 if self.ethnicity == BLACK else 1.000

    @property
    def gender_factor(self) -> float:
        return 1.018 if self.gender == FEMALE else 1.000

    @property
    def age_factor(self) -> float:
        return 0.993**self.age_in_years



import sys
from typing import Optional

from django.core.management.base import BaseCommand
from django.core.management.color import color_style
from edc_protocol.protocol import Protocol

from edc_reportable.site_reportables import site_reportables

style = color_style()


def export_daids_grading(path: Optional[str]):
    sys.stdout.write(style.MIGRATE_HEADING("Exporting reportables to document (.csv) ...\n"))
    collection_name = Protocol().project_name.lower()
    filename1, filename2 = site_reportables.to_csv(collection_name=collection_name, path=path)
    sys.stdout.write(
        style.MIGRATE_HEADING(f"  * Exported to `{filename1}` and `{filename2}`\n")
    )
    sys.stdout.write(style.MIGRATE_HEADING("Done\n"))


class Command(BaseCommand):

    help = "Export export DAIDS grading to document (.csv)"

    def add_arguments(self, parser):
        parser.add_argument(
            "--path",
            default="~/",
            action="store_true",
            dest="path",
            help="Export path/folder",
        )

    def handle(self, *args, **options):

        export_daids_grading(path=options["path"])



from datetime import datetime

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import MALE
from edc_utils import age
from pytz import utc

from edc_reportable import (
    AgeEvaluator,
    Evaluator,
    InvalidCombination,
    InvalidLowerBound,
    InvalidUnits,
    InvalidUpperBound,
    NormalReference,
    ValueBoundryError,
)


class TestEvaluators(TestCase):
    def test_evaluator_zero(self):
        """Test the basic evaluator."""
        ref = Evaluator(
            lower=0,
            upper=100,
            units="mg/dL",
            lower_inclusive=False,
            upper_inclusive=False,
        )

        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 0, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(0.1, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99.9, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 100, units="mg/dL")

    def test_evaluator_lower_none(self):
        """Test the basic evaluator."""
        ref = Evaluator(upper=100, units="mg/dL", upper_inclusive=False)

        self.assertTrue(ref.in_bounds_or_raise(-1, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(0.0, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(0.1, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99.9, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 100, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 10000, units="mg/dL")

    def test_evaluator_upper_none(self):
        """Test the basic evaluator."""
        ref = Evaluator(lower=100, units="mg/dL", lower_inclusive=True)

        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 0, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 99, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(100, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(10000, units="mg/dL"))

    def test_evaluator(self):
        """Test the basic evaluator."""

        ref = Evaluator(lower=10, upper=100, units="mg/dL")
        self.assertTrue(repr(ref))
        self.assertTrue(str(ref))

        self.assertRaises(InvalidUnits, Evaluator, lower=10, upper=100, units=None)

        ref = Evaluator(lower=10, upper=100, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 9, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 10, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(11, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 100, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 101, units="mg/dL")

        ref = Evaluator(lower=10, upper=100, units="mg/dL", lower_inclusive=True)
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 9, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(10, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(11, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 100, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 101, units="mg/dL")

        ref = Evaluator(
            lower=10,
            upper=100,
            units="mg/dL",
            lower_inclusive=True,
            upper_inclusive=True,
        )
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 9, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(10, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(11, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(100, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 101, units="mg/dL")

        ref = Evaluator(lower=10, upper=100, units="mg/dL", upper_inclusive=True)
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 9, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 10, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(11, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(100, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 101, units="mg/dL")

        ref = Evaluator(lower=10, upper=None, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 9, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 10, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(11, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(100, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(100000000000, units="mg/dL"))

        ref = Evaluator(lower=None, upper=100, units="mg/dL")
        self.assertTrue(ref.in_bounds_or_raise(-1, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(1, units="mg/dL"))
        self.assertTrue(ref.in_bounds_or_raise(99, units="mg/dL"))
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 100, units="mg/dL")
        self.assertRaises(ValueBoundryError, ref.in_bounds_or_raise, 101, units="mg/dL")

        self.assertRaises(InvalidUnits, ref.in_bounds_or_raise, 101, units="blah")

        ref = Evaluator(lower=None, upper=100, units="mg/dL")
        self.assertEqual(ref.description(), "x<100.0 mg/dL")
        self.assertRaises(InvalidLowerBound, Evaluator, lower="ERIK", upper=100, units="mg/dL")
        self.assertRaises(InvalidUpperBound, Evaluator, lower=10, upper="ERIK", units="mg/dL")
        ref = Evaluator(lower=10, upper=None, units="mg/dL")
        self.assertEqual(ref.description(), "10.0<x mg/dL")

        for lower in [0.1, 1.1, 10.2234]:
            with self.subTest(lower=lower):
                try:
                    Evaluator(lower=lower, upper=100, units="mg/dL")
                except InvalidLowerBound:
                    self.fail("InvalidLowerBound unexpectedly raised")

        for upper in [0.5, 1.1, 10.2234]:
            with self.subTest(upper=upper):
                try:
                    Evaluator(lower=0.1, upper=upper, units="mg/dL")
                except InvalidUpperBound:
                    self.fail("InvalidUpperBound unexpectedly raised")

        self.assertRaises(InvalidCombination, Evaluator, lower=10, upper=10, units="mg/dL")
        self.assertRaises(InvalidCombination, Evaluator, lower=11, upper=10, units="mg/dL")

    def test_age_evaluator(self):
        """Test the age evaluator which is a child class
        of the basic evaluator.
        """
        report_datetime = utc.localize(datetime(2017, 12, 7))
        dob = report_datetime - relativedelta(years=25)
        rdelta = age(dob, report_datetime)

        self.assertEqual(age(dob, report_datetime).years, 25)
        self.assertTrue(24 < getattr(rdelta, "years") < 26)
        self.assertFalse(25 < getattr(rdelta, "years") < 26)
        self.assertFalse(24 < getattr(rdelta, "years") < 25)

        age_eval = AgeEvaluator(age_lower=24, age_upper=26)
        self.assertTrue(repr(age_eval))
        self.assertTrue(age_eval.in_bounds_or_raise(dob, report_datetime))
        self.assertEqual(age_eval.description(), "24<AGE<26 years")

        age_eval = AgeEvaluator(age_lower=18)
        self.assertTrue(repr(age_eval))
        self.assertTrue(age_eval.in_bounds_or_raise(dob, report_datetime))
        self.assertEqual(age_eval.description(), "18<AGE years")

        age_eval = AgeEvaluator(age_lower=25, age_upper=26)
        self.assertRaises(ValueBoundryError, age_eval.in_bounds_or_raise, dob, report_datetime)
        self.assertEqual(age_eval.description(), "25<AGE<26 years")

        age_eval = AgeEvaluator(age_lower=24, age_upper=25)
        self.assertRaises(ValueBoundryError, age_eval.in_bounds_or_raise, dob, report_datetime)
        self.assertEqual(age_eval.description(), "24<AGE<25 years")

    def test_age_match(self):
        """Test age within the NormalReference."""
        report_datetime = utc.localize(datetime(2017, 12, 7))
        dob = report_datetime - relativedelta(years=25)

        ref = NormalReference(
            lower=10, upper=None, units="mg/dL", age_lower=24, age_upper=26, gender=MALE
        )
        self.assertTrue(ref.age_match(dob, report_datetime))

        ref = NormalReference(
            lower=10, upper=None, units="mg/dL", age_lower=25, age_upper=26, gender=MALE
        )
        self.assertFalse(ref.age_match(dob, report_datetime))

        ref = NormalReference(
            lower=10, upper=None, units="mg/dL", age_lower=24, age_upper=25, gender=MALE
        )
        self.assertFalse(ref.age_match(dob, report_datetime))



from django.db import models
from django.db.models.deletion import CASCADE
from edc_constants.choices import YES_NO, YES_NO_NA
from edc_model.models import BaseUuidModel
from edc_utils import get_utcnow

from edc_reportable.choices import REPORTABLE
from edc_reportable.units import GRAMS_PER_DECILITER


class SubjectVisit(BaseUuidModel):

    subject_identifier = models.CharField(max_length=25)


class SpecimenResult(BaseUuidModel):

    subject_visit = models.ForeignKey(SubjectVisit, on_delete=CASCADE)

    report_datetime = models.DateTimeField(default=get_utcnow)

    haemoglobin = models.DecimalField(decimal_places=1, max_digits=6, null=True, blank=True)

    haemoglobin_units = models.CharField(
        verbose_name="units",
        max_length=10,
        choices=((GRAMS_PER_DECILITER, GRAMS_PER_DECILITER),),
        null=True,
        blank=True,
    )

    haemoglobin_abnormal = models.CharField(
        verbose_name="abnormal", choices=YES_NO, max_length=25, null=True, blank=True
    )

    haemoglobin_reportable = models.CharField(
        verbose_name="reportable",
        choices=REPORTABLE,
        max_length=25,
        null=True,
        blank=True,
    )

    results_abnormal = models.CharField(
        verbose_name="Are any of the above results abnormal?",
        choices=YES_NO,
        max_length=25,
    )

    results_reportable = models.CharField(
        verbose_name="If any results are abnormal, are results within grade III " "or above?",
        max_length=25,
        choices=YES_NO_NA,
    )

    @property
    def abnormal(self):
        return self.results_abnormal

    @property
    def reportable(self):
        return self.results_reportable



from datetime import datetime

from dateutil.relativedelta import relativedelta
from django.test import TestCase
from edc_constants.constants import FEMALE, MALE
from edc_utils import get_utcnow
from pytz import utc

from edc_reportable import (
    BoundariesOverlap,
    InvalidValueReference,
    NormalReference,
    NormalReferenceError,
    NotEvaluated,
    ValueReferenceAlreadyAdded,
    ValueReferenceGroup,
)
from edc_reportable.normal_data.africa import normal_data
from edc_reportable.units import IU_LITER, MILLIMOLES_PER_LITER


class TestValueReference(TestCase):
    def test_value_reference_group(self):
        dob = get_utcnow() - relativedelta(years=25)
        report_datetime = utc.localize(datetime(2017, 12, 7))
        grp = ValueReferenceGroup(name="labtest")
        self.assertTrue(repr(grp))

        ref = NormalReference(
            name="blahblah",
            lower=10,
            upper=None,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=MALE,
        )

        self.assertRaises(InvalidValueReference, grp.add_normal, ref)

        ref = NormalReference(
            name="labtest",
            lower=10,
            upper=None,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=MALE,
        )

        grp.add_normal(ref)

        self.assertFalse(grp.get_normal(value=9, units="mg/dL", gender=MALE, dob=dob))
        self.assertFalse(grp.get_normal(value=10, units="mg/dL", gender=MALE, dob=dob))
        self.assertTrue(grp.get_normal(value=11, units="mg/dL", gender=MALE, dob=dob))
        self.assertRaises(ValueReferenceAlreadyAdded, grp.add_normal, ref)

        # try without upper bound age
        grp = ValueReferenceGroup(name="another_labtest")
        ref = NormalReference(
            name="another_labtest",
            lower=10,
            upper=None,
            units="mg/dL",
            age_lower=18,
            age_units="years",
            gender=MALE,
        )

        grp.add_normal(ref)

        self.assertFalse(grp.get_normal(value=9, units="mg/dL", gender=MALE, dob=dob))
        self.assertFalse(grp.get_normal(value=10, units="mg/dL", gender=MALE, dob=dob))
        self.assertTrue(grp.get_normal(value=11, units="mg/dL", gender=MALE, dob=dob))
        self.assertRaises(ValueReferenceAlreadyAdded, grp.add_normal, ref)
        self.assertEqual(
            grp.get_normal_description(units="mg/dL", gender=MALE, dob=dob),
            ["10.0<x mg/dL M 18<AGE years"],
        )

        grp = ValueReferenceGroup(name="labtest")

        ref_male = NormalReference(
            name="labtest",
            lower=10,
            upper=None,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=MALE,
        )
        ref_female1 = NormalReference(
            name="labtest",
            lower=1.7,
            upper=3.5,
            upper_inclusive=True,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=FEMALE,
        )
        ref_female2 = NormalReference(
            name="labtest",
            lower=7.3,
            upper=None,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=FEMALE,
        )

        grp.add_normal(ref_male)
        grp.add_normal(ref_female1)
        grp.add_normal(ref_female2)

        self.assertFalse(
            grp.get_normal(
                value=9,
                gender=MALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertFalse(
            grp.get_normal(
                value=10,
                gender=MALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertTrue(
            grp.get_normal(
                value=11,
                gender=MALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )

        self.assertFalse(
            grp.get_normal(
                value=1,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertFalse(
            grp.get_normal(
                value=1.7,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertTrue(
            grp.get_normal(
                value=1.8,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertTrue(
            grp.get_normal(
                value=3.4,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertTrue(
            grp.get_normal(
                value=3.5,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertFalse(
            grp.get_normal(
                value=3.6,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )

        self.assertFalse(
            grp.get_normal(
                value=7.3,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertTrue(
            grp.get_normal(
                value=7.4,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )

        self.assertRaises(
            NotEvaluated,
            grp.get_normal,
            value=7.4,
            gender=FEMALE,
            dob=report_datetime.date(),
            report_datetime=report_datetime,
            units="mg/dL",
        )

        self.assertRaises(
            NotEvaluated,
            grp.get_normal,
            value=7.4,
            gender=FEMALE,
            dob=report_datetime.date(),
            report_datetime=report_datetime,
            units="mmol/L",
        )

        # for a normal value, show what it was evaluated against
        # for messaging
        self.assertFalse(
            grp.get_normal(
                value=7.3,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units="mg/dL",
            )
        )
        self.assertEqual(
            grp.get_normal_description(
                gender=FEMALE, dob=dob, report_datetime=report_datetime, units="mg/dL"
            ),
            ["1.7<x<=3.5 mg/dL F 18<AGE<99 years", "7.3<x mg/dL F 18<AGE<99 years"],
        )

        # overlaps with ref_female3
        ref_female4 = NormalReference(
            name="labtest",
            lower=7.3,
            upper=9.3,
            units="mg/dL",
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=FEMALE,
        )
        grp.add_normal(ref_female4)

        self.assertRaises(
            BoundariesOverlap,
            grp.get_normal,
            value=7.4,
            gender=FEMALE,
            dob=dob,
            report_datetime=report_datetime,
            units="mg/dL",
        )

        # adds extra attributes
        grp = ValueReferenceGroup(name="glucose")
        ref_glu = NormalReference(
            name="glucose",
            lower=4.8,
            upper=5.6,
            units=MILLIMOLES_PER_LITER,
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=[MALE, FEMALE],
            fasting=True,
        )
        grp.add_normal(ref_glu)

        self.assertTrue(
            grp.get_normal(
                value=5.4,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units=MILLIMOLES_PER_LITER,
                fasting=True,
            )
        )

        self.assertFalse(
            grp.get_normal(
                value=7.4,
                gender=FEMALE,
                dob=dob,
                report_datetime=report_datetime,
                units=MILLIMOLES_PER_LITER,
                fasting=True,
            )
        )

        self.assertRaises(
            NotEvaluated,
            grp.get_normal,
            value=5.4,
            gender=FEMALE,
            dob=dob,
            report_datetime=report_datetime,
            units=MILLIMOLES_PER_LITER,
            fasting=False,
        )

    def test_value_reference_group_with_ll(self):
        opts = dict(
            name="amylase",
            lower="1.5*ULN",
            upper="3.0*ULN",
            units=IU_LITER,
            age_lower=18,
            age_upper=99,
            age_units="years",
            gender=[MALE, FEMALE],
            fasting=True,
            normal_references={"MF": normal_data.get("amylase")},
        )

        self.assertRaises(NormalReferenceError, NormalReference, **opts)



from .adult_age_options import adult_age_options
from .age_evaluator import AgeEvaluator
from .calculators import (
    BMI,
    CalculatorError,
    EgfrCkdEpi,
    EgfrCockcroftGault,
    calculate_bmi,
)
from .constants import (
    ALREADY_REPORTED,
    GRADE0,
    GRADE1,
    GRADE2,
    GRADE3,
    GRADE4,
    GRADE5,
    MILD,
    MODERATE,
    PRESENT_AT_BASELINE,
    SEVERE,
    SEVERITY_INCREASED_FROM_G3,
)
from .convert_units import ConversionNotHandled, convert_units
from .evaluator import (
    Evaluator,
    InvalidCombination,
    InvalidLowerBound,
    InvalidUnits,
    InvalidUpperBound,
    ValueBoundryError,
)
from .form_validator_mixins import (
    BmiFormValidatorMixin,
    EgfrCkdEpiFormValidatorMixin,
    EgfrCockcroftGaultFormValidatorMixin,
    ReportablesFormValidatorMixin,
)
from .grade_reference import GradeError, GradeReference
from .normal_reference import NormalReference, NormalReferenceError
from .parsers import ParserError, dummy_parse, parse, unparse
from .reference_range_collection import AlreadyRegistered, ReferenceRangeCollection
from .reportables_evaluator import ReportablesEvaluator
from .site_reportables import site_reportables
from .units import (
    CELLS_PER_MICROLITER,
    CELLS_PER_MILLIMETER_CUBED,
    CELLS_PER_MILLIMETER_CUBED_DISPLAY,
    COPIES_PER_MILLILITER,
    GRAMS_PER_DECILITER,
    GRAMS_PER_LITER,
    IU_LITER,
    IU_LITER_DISPLAY,
    MICROMOLES_PER_LITER,
    MICROMOLES_PER_LITER_DISPLAY,
    MILLIGRAMS_PER_DECILITER,
    MILLILITER_PER_MINUTE,
    MILLIMOLES_PER_LITER,
    MILLIMOLES_PER_LITER_DISPLAY,
    MM3,
    MM3_DISPLAY,
    PERCENT,
    TEN_X_3_PER_LITER,
    TEN_X_3_PER_LITER_DISPLAY,
    TEN_X_9_PER_LITER,
    TEN_X_9_PER_LITER_DISPLAY,
)
from .value_reference_group import (
    BoundariesOverlap,
    InvalidValueReference,
    NotEvaluated,
    ValueReferenceAlreadyAdded,
    ValueReferenceGroup,
)



import sys

from django.apps import AppConfig as BaseAppConfig

from .site_reportables import site_reportables


class AppConfig(BaseAppConfig):
    name = "edc_reportable"

    def ready(self):
        sys.stdout.write(f"Loading {self.verbose_name} ...\n")
        site_reportables.autodiscover()
        sys.stdout.write(f" Done loading {self.verbose_name}.\n")



# from django.db import models
# from edc_model.models import BaseUuidModel
#
#
# class LimitNormal(BaseUuidModel):
#     name = models.CharField(max_length=25,)
#
#     lln_factor = models.IntegerField(default=1)
#
#     uln_factor = models.IntegerField(default=1)
#
#     class Meta:
#         constraints = ()



from django.test import TestCase
from edc_constants.constants import MALE

from edc_reportable import ParserError, parse, unparse


class TestParser(TestCase):
    def test1(self):
        p = parse("7<x<8")
        self.assertEqual(p.lower, 7)
        self.assertFalse(p.lower_inclusive)
        self.assertEqual(p.upper, 8)
        self.assertFalse(p.upper_inclusive)

    def test2(self):
        p = parse("7<=x<8")
        self.assertEqual(p.lower, 7)
        self.assertTrue(p.lower_inclusive)
        self.assertEqual(p.upper, 8)
        self.assertFalse(p.upper_inclusive)

    def test3(self):
        p = parse("7<x<=8")
        self.assertEqual(p.lower, 7)
        self.assertFalse(p.lower_inclusive)
        self.assertEqual(p.upper, 8)
        self.assertTrue(p.upper_inclusive)

    def test4(self):
        p = parse("7<=x<=8")
        self.assertEqual(p.lower, 7)
        self.assertTrue(p.lower_inclusive)
        self.assertEqual(p.upper, 8)
        self.assertTrue(p.upper_inclusive)

    def test5(self):
        p = parse(".7<=x<=.8")
        self.assertEqual(p.lower, 0.7)
        self.assertTrue(p.lower_inclusive)
        self.assertEqual(p.upper, 0.8)
        self.assertTrue(p.upper_inclusive)

    def test6(self):
        p = parse("0.77<=x<=0.88")
        self.assertEqual(p.lower, 0.77)
        self.assertTrue(p.lower_inclusive)
        self.assertEqual(p.upper, 0.88)
        self.assertTrue(p.upper_inclusive)

    def test7(self):
        p = parse("0.77 <= x <= 0.88")
        self.assertEqual(p.lower, 0.77)
        self.assertTrue(p.lower_inclusive)
        self.assertEqual(p.upper, 0.88)
        self.assertTrue(p.upper_inclusive)

    def test8(self):
        p = parse("x <= 0.88")
        self.assertIsNone(p.lower)
        self.assertIsNone(p.lower_inclusive)
        self.assertEqual(p.upper, 0.88)
        self.assertTrue(p.upper_inclusive)

    def test9(self):
        p = parse("0.77 <= x")
        self.assertEqual(p.lower, 0.77)
        self.assertTrue(p.lower_inclusive)
        self.assertIsNone(p.upper)
        self.assertIsNone(p.upper_inclusive)

    def test10(self):
        p = parse("0.77 <= x <= 0.88")
        self.assertEqual(unparse(**p), "0.77<=x<=0.88")
        p = parse("0.77 <= x <= 0.88")
        self.assertEqual(unparse(gender=MALE, **p), "0.77<=x<=0.88 M")

    def test11(self):
        self.assertRaises(ParserError, parse, "0.77 <= x = 0.88")
        self.assertRaises(ParserError, parse, "0.77 <= x =")

        self.assertRaises(ParserError, parse, "<0.77")

        self.assertRaises(ParserError, parse, "<77")

        self.assertRaises(ParserError, parse, "=77")

        self.assertRaises(ParserError, parse, ">77")

        self.assertRaises(ParserError, parse, "0.77 >= x > 0.88")

        self.assertRaises(ParserError, parse, "0.77 =< x < 0.88")

        self.assertRaises(ParserError, parse, "0.77 < x < 0.88 < x < 0.88")
